from .abc import AbstractFactory, AbstractConnection
from .connection import WebsocketPool, AiohttpFactory
from .client import GremlinClient
from .exceptions import RequestError, GremlinServerError, SocketClientError
__version__ = "0.0.1"



"""
"""
import asyncio

import aiohttp

from aiogremlin.abc import AbstractFactory, AbstractConnection
from aiogremlin.exceptions import SocketClientError
from aiogremlin.log import INFO, conn_logger


class WebsocketPool:

    def __init__(self, uri='ws://localhost:8182/', factory=None, poolsize=10,
                 max_retries=10, timeout=None, loop=None, verbose=False):
        """
        """
        self.uri = uri
        self._factory = factory or AiohttpFactory
        self.poolsize = poolsize
        self.max_retries = max_retries
        self.timeout = timeout
        self._loop = loop or asyncio.get_event_loop()
        self.pool = asyncio.Queue(maxsize=self.poolsize, loop=self._loop)
        self.active_conns = set()
        self.num_connecting = 0
        self._closed = False
        if verbose:
            conn_logger.setLevel(INFO)

    @property
    def loop(self):
        return self._loop

    @property
    def factory(self):
        return self._factory

    @property
    def closed(self):
        return self._closed

    @property
    def num_active_conns(self):
        return len(self.active_conns)

    def feed_pool(self, conn):
        if self._closed:
            raise RuntimeError("WebsocketPool is closed.")
        self.active_conns.discard(conn)
        self._put(conn)

    @asyncio.coroutine
    def close(self):
        if not self._closed:
            if self.active_conns:
                yield from self._close_active_conns()
            yield from self._purge_pool()
            self._closed = True

    @asyncio.coroutine
    def _close_active_conns(self):
        tasks = [asyncio.async(conn.close(), loop=self.loop) for conn
            in self.active_conns]
        yield from asyncio.wait(tasks, loop=self.loop)

    @asyncio.coroutine
    def _purge_pool(self):
        while True:
            try:
                conn = self.pool.get_nowait()
            except asyncio.QueueEmpty:
                break
            else:
                yield from conn.close()

    @asyncio.coroutine
    def connect(self, uri=None, loop=None, num_retries=None):
        if self._closed:
            raise RuntimeError("WebsocketPool is closed.")
        if num_retries is None:
            num_retries = self.max_retries
        uri = uri or self.uri
        loop = loop or self.loop
        if not self.pool.empty():
            socket = self.pool.get_nowait()
            conn_logger.info("Reusing socket: {} at {}".format(socket, uri))
        elif (self.num_active_conns + self.num_connecting >= self.poolsize or
            not self.poolsize):
            conn_logger.info("Waiting for socket...")
            socket = yield from asyncio.wait_for(self.pool.get(),
                self.timeout, loop=loop)
            conn_logger.info("Socket acquired: {} at {}".format(socket, uri))
        else:
            self.num_connecting += 1
            try:
                socket = yield from self.factory.connect(uri, pool=self,
                    loop=loop)
            except:
                raise
            else:
                conn_logger.info("New connection on socket: {} at {}".format(
                    socket, uri))
            finally:
                self.num_connecting -= 1
        if not socket.closed:
            self.active_conns.add(socket)
        # Untested.
        elif num_retries > 0:
            socket = yield from self.connect(uri, loop, num_retries - 1)
        else:
            raise RuntimeError("Unable to connect, max retries exceeded.")
        return socket

    def _put(self, socket):
        try:
            self.pool.put_nowait(socket)
        except asyncio.QueueFull:
            pass
            # This should be - not working
            # yield from socket.release()


class BaseFactory(AbstractFactory):

    @property
    def factory(self):
        return self


class AiohttpFactory(BaseFactory):

    @classmethod
    @asyncio.coroutine
    def connect(cls, uri='ws://localhost:8182/', pool=None, protocols=(),
                connector=None, autoclose=False, autoping=False, loop=None):
        if pool:
            loop = loop or pool.loop
        try:
            socket = yield from aiohttp.ws_connect(uri, protocols=protocols,
                connector=connector, autoclose=autoclose, autoping=autoping,
                loop=loop)
        except aiohttp.WSServerHandshakeError as e:
            raise SocketClientError(e.message)
        return AiohttpConnection(socket, pool)


class BaseConnection(AbstractConnection):

    def __init__(self, socket, pool=None):
        self.socket = socket
        self._pool = pool

    def feed_pool(self):
        if self.pool:
            if self in self.pool.active_conns:
                self.pool.feed_pool(self)

    @asyncio.coroutine
    def receive(self):
        return (yield from parse_gremlin_response(self))

    @asyncio.coroutine
    def release(self):
        try:
            yield from self.close()
        finally:
            if self in self.pool.active_conns:
                self.pool.active_conns.discard(self)

    @property
    def pool(self):
        return self._pool


class AiohttpConnection(BaseConnection):

    @property
    def closed(self):
        return self.socket.closed

    @asyncio.coroutine
    def close(self):
        if not self.socket._closed:
            try:
                yield from self.socket.close()
            finally:
                # Socket should close despite errors.
                self._closed = True

    @asyncio.coroutine
    def send(self, message, binary=True):
        if binary:
            method = self.socket.send_bytes
        else:
            method = self.socket.send_str
        try:
            method(message)
        except RuntimeError:
            # Socket closed.
            yield from self.release()
            raise
        except TypeError:
            # Bytes/string input error.
            yield from self.release()
            raise

    @asyncio.coroutine
    def _receive(self):
        """Implements a dispatcher using the aiohttp websocket protocol."""
        while True:
            try:
                message = yield from self.socket.receive()
            except (asyncio.CancelledError, asyncio.TimeoutError):
                yield from self.release()
                raise
            except RuntimeError:
                yield from self.release()
                raise
            if message.tp == aiohttp.MsgType.binary:
                return message.data.decode()
            elif message.tp == aiohttp.MsgType.text:
                return message.data.strip()
            elif message.tp == aiohttp.MsgType.ping:
                conn_logger.warn("Ping received.")
                ws.pong()
                conn_logger.warn("Sent pong.")
            elif message.tp == aiohttp.MsgType.pong:
                conn_logger.warn('Pong received')
            else:
                try:
                    if message.tp == aiohttp.MsgType.release:
                        conn_logger.warn("Socket connection closed by server.")
                    elif message.tp == aiohttp.MsgType.error:
                        raise SocketClientError(self.socket.exception())
                    elif message.tp == aiohttp.MsgType.closed:
                        raise RuntimeError("Socket closed.")
                    break
                finally:
                    yield from self.release()



"""
"""

import asyncio
import itertools
import unittest
from aiogremlin import (GremlinClient, RequestError, GremlinServerError,
    SocketClientError, WebsocketPool, AiohttpFactory)


class GremlinClientTests(unittest.TestCase):

    def setUp(self):
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(None)
        self.gc = GremlinClient("ws://localhost:8182/",
            factory=AiohttpFactory, loop=self.loop)

    def tearDown(self):
        self.loop.run_until_complete(self.gc.close())
        self.loop.close()

    def test_connection(self):
        @asyncio.coroutine
        def conn_coro():
            conn = yield from self.gc.connect()
            self.assertFalse(conn.closed)
            return conn
        conn = self.loop.run_until_complete(conn_coro())
        # Clean up the resource.
        self.loop.run_until_complete(conn.close())

    def test_sub(self):
        execute = self.gc.execute("x + x", bindings={"x": 4})
        results = self.loop.run_until_complete(execute)
        self.assertEqual(results[0].data[0], 8)

    def test_sub_waitfor(self):
        sub1 = self.gc.execute("x + x", bindings={"x": 1})
        sub2 = self.gc.execute("x + x", bindings={"x": 2})
        sub3 = self.gc.execute("x + x", bindings={"x": 4})
        coro = asyncio.wait([asyncio.async(sub1, loop=self.loop),
                             asyncio.async(sub2, loop=self.loop),
                             asyncio.async(sub3, loop=self.loop)], loop=self.loop)
        # Here I am looking for resource warnings.
        results = self.loop.run_until_complete(coro)
        self.assertIsNotNone(results)

    def test_resp_stream(self):
        @asyncio.coroutine
        def stream_coro():
            results = []
            resp = yield from self.gc.submit("x + x", bindings={"x": 4})
            while True:
                f = yield from resp.stream.read()
                if f is None:
                    break
                results.append(f)
            self.assertEqual(results[0].data[0], 8)
        self.loop.run_until_complete(stream_coro())

    def test_resp_get(self):
        @asyncio.coroutine
        def get_coro():
            conn = yield from self.gc.submit("x + x", bindings={"x": 4})
            results = yield from conn.get()
            self.assertEqual(results[0].data[0], 8)
        self.loop.run_until_complete(get_coro())

    def test_execute_error(self):
        execute = self.gc.execute("x + x g.asdfas", bindings={"x": 4})
        try:
            self.loop.run_until_complete(execute)
            error = False
        except:
            error = True
        self.assertTrue(error)


class WebsocketPoolTests(unittest.TestCase):

    def setUp(self):
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(None)
        self.pool = WebsocketPool(poolsize=2, timeout=1, loop=self.loop,
            factory=AiohttpFactory)

    def tearDown(self):
        self.loop.run_until_complete(self.pool.close())
        self.loop.close()

    def test_connect(self):

        @asyncio.coroutine
        def conn():
            conn = yield from self.pool.connect()
            self.assertIsNotNone(conn.socket)
            self.assertFalse(conn.closed)
            conn.feed_pool()
            self.assertEqual(self.pool.num_active_conns, 0)

        self.loop.run_until_complete(conn())

    def test_multi_connect(self):

        @asyncio.coroutine
        def conn():
            conn1 = yield from self.pool.connect()
            conn2 = yield from self.pool.connect()
            self.assertIsNotNone(conn1.socket)
            self.assertFalse(conn1.closed)
            self.assertIsNotNone(conn2.socket)
            self.assertFalse(conn2.closed)
            conn1.feed_pool()
            self.assertEqual(self.pool.num_active_conns, 1)
            conn2.feed_pool()
            self.assertEqual(self.pool.num_active_conns, 0)

        self.loop.run_until_complete(conn())

    def test_timeout(self):

        @asyncio.coroutine
        def conn():
            conn1 = yield from self.pool.connect()
            conn2 = yield from self.pool.connect()
            try:
                conn3 = yield from self.pool.connect()
                timeout = False
            except asyncio.TimeoutError:
                timeout = True
            self.assertTrue(timeout)

        self.loop.run_until_complete(conn())

    def test_socket_reuse(self):

        @asyncio.coroutine
        def conn():
            conn1 = yield from self.pool.connect()
            conn2 = yield from self.pool.connect()
            try:
                conn3 = yield from self.pool.connect()
                timeout = False
            except asyncio.TimeoutError:
                timeout = True
            self.assertTrue(timeout)
            conn2.feed_pool()
            conn3 = yield from self.pool.connect()
            self.assertIsNotNone(conn1.socket)
            self.assertFalse(conn1.closed)
            self.assertIsNotNone(conn3.socket)
            self.assertFalse(conn3.closed)
            self.assertEqual(conn2.socket, conn3.socket)

        self.loop.run_until_complete(conn())

    def test_socket_repare(self):

        @asyncio.coroutine
        def conn():
            conn1 = yield from self.pool.connect()
            conn2 = yield from self.pool.connect()
            self.assertIsNotNone(conn1.socket)
            self.assertFalse(conn1.closed)
            self.assertIsNotNone(conn2.socket)
            self.assertFalse(conn2.closed)
            yield from conn1.socket.close()
            yield from conn2.socket.close()
            self.assertTrue(conn2.closed)
            self.assertTrue(conn2.closed)
            conn1.feed_pool()
            conn2.feed_pool()
            conn1 = yield from self.pool.connect()
            conn2 = yield from self.pool.connect()
            self.assertIsNotNone(conn1.socket)
            self.assertFalse(conn1.closed)
            self.assertIsNotNone(conn2.socket)
            self.assertFalse(conn2.closed)

        self.loop.run_until_complete(conn())


if __name__ == "__main__":
    unittest.main()



import logging


INFO = logging.INFO


logging.basicConfig(
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s")


client_logger = logging.getLogger("aiogremlin.client")
conn_logger = logging.getLogger("aiogremlin.connection")



"""Implements a very simple "protocol" for the Gremlin server."""

import asyncio
import collections
import json

from aiogremlin.exceptions import RequestError, GremlinServerError


Message = collections.namedtuple("Message", ["status_code", "data", "message",
    "metadata"])


@asyncio.coroutine
def gremlin_response_parser(connection):
    message = yield from connection._receive()
    message = json.loads(message)
    message = Message(message["status"]["code"],
                      message["result"]["data"],
                      message["result"]["meta"],
                      message["status"]["message"])
    if message.status_code == 200:
        return message
    elif message.status_code == 299:
        connection.feed_pool()
        # Return None
    else:
        try:
            if message.status_code < 500:
                raise RequestError(message.status_code, message.message)
            else:
                raise GremlinServerError(message.status_code, message.message)
        finally:
            yield from connection.release()


class GremlinWriter:

    def __init__(self, connection):
        self._connection = connection

    @asyncio.coroutine
    def write(self, message, binary=True, mime_type="application/json"):
        if binary:
            message = self._set_message_header(message, mime_type)
        yield from self._connection.send(message, binary)
        return self._connection

    @staticmethod
    def _set_message_header(message, mime_type):
        if mime_type == "application/json":
            mime_len = b"\x10"
            mime_type = b"application/json"
        else:
            raise ValueError("Unknown mime type.")
        return b"".join([mime_len, mime_type, bytes(message, "utf-8")])



"""Client for the Tinkerpop 3 Gremlin Server."""

import asyncio
import json
import ssl
import uuid

from aiogremlin.connection import WebsocketPool
from aiogremlin.log import client_logger
from aiogremlin.protocol import gremlin_response_parser, GremlinWriter


class GremlinClient:

    def __init__(self, uri='ws://localhost:8182/', loop=None, ssl=None,
                 protocol=None, lang="gremlin-groovy", op="eval",
                 processor="", pool=None, factory=None, poolsize=10,
                 timeout=None, **kwargs):
        """
        """
        self.uri = uri
        self.ssl = ssl
        self.protocol = protocol
        # if self.ssl:
        #     protocol = protocol or ssl.PROTOCOL_TLSv1
        #     ssl_context = ssl.SSLContext(protocol)
        #     ssl_context.load_verify_locations(ssl)
        #     ssl_context.verify_mode = ssl.CERT_REQUIRED
        #     self.ssl_context = ssl_context  # This will go to conn pool... use TCPConnector?
        self._loop = loop or asyncio.get_event_loop()
        self.lang = lang or "gremlin-groovy"
        self.op = op or "eval"
        self.processor = processor or ""
        self.poolsize = poolsize
        self.timeout = timeout
        self.pool = pool or WebsocketPool(uri, factory=factory,
            poolsize=poolsize, timeout=timeout, loop=self._loop)
        self.factory = factory or self.pool.factory

    @property
    def loop(self):
        return self._loop

    @asyncio.coroutine
    def close(self):
        yield from self.pool.close()

    @asyncio.coroutine
    def connect(self, **kwargs):
        """
        """
        loop = kwargs.get("loop", "") or self.loop
        connection = yield from self.factory.connect(self.uri, loop=loop,
            **kwargs)
        return connection

    @asyncio.coroutine
    def submit(self, gremlin, connection=None, bindings=None, lang=None,
               op=None, processor=None, binary=True):
        """
        """
        lang = lang or self.lang
        op = op or self.op
        processor = processor or self.processor
        message = json.dumps({
            "requestId": str(uuid.uuid4()),
            "op": op,
            "processor": processor,
            "args":{
                "gremlin": gremlin,
                "bindings": bindings,
                "language":  lang
            }
        })
        if connection is None:
            connection = yield from self.pool.connect(self.uri, loop=self.loop)
        writer = GremlinWriter(connection)
        connection = yield from writer.write(message, binary=binary)
        return GremlinResponse(connection)

    @asyncio.coroutine
    def execute(self, gremlin, bindings=None, lang=None,
               op=None, processor=None, consumer=None, collect=True, **kwargs):
        """
        """
        lang = lang or self.lang
        op = op or self.op
        processor = processor or self.processor
        resp = yield from self.submit(gremlin, bindings=bindings, lang=lang,
                op=op, processor=processor)
        return (yield from resp.get())


class GremlinResponse:

    def __init__(self, conn):
        self._stream = GremlinResponseStream(conn)

    @property
    def stream(self):
        return self._stream

    @asyncio.coroutine
    def get(self):
        return (yield from self._run())

    @asyncio.coroutine
    def _run(self):
        """
        """
        results = []
        while True:
            message = yield from self.stream.read()
            if message is None:
                break
            results.append(message)
        return results


class GremlinResponseStream:

    def __init__(self, conn):
        self.conn = conn

    @asyncio.coroutine
    def read(self):
        return (yield from gremlin_response_parser(self.conn))



"""Abstract classes for creating pluggable websocket clients."""

from abc import ABCMeta, abstractmethod


class AbstractFactory(metaclass=ABCMeta):

    @classmethod
    @abstractmethod
    def connect(cls):
        pass

    @property
    @abstractmethod
    def factory(self):
        pass


class AbstractConnection(metaclass=ABCMeta):

    @abstractmethod
    def feed_pool(self):
        pass

    @abstractmethod
    def release(self):
        pass

    @property
    @abstractmethod
    def pool(self):
        pass

    @property
    @abstractmethod
    def closed(self):
        pass

    @abstractmethod
    def close():
        pass

    @abstractmethod
    def send(self):
        pass

    @abstractmethod
    def receive(self):
        pass
